typedef struct Attr Attr;
typedef struct Node Node;
typedef struct HtmlParser HtmlParser;

struct Attr
{
	Rune *name;
	Rune *value;
};

struct Node
{
	int		id;
	int		type;
	union{
		struct{				/* Nelement */
			int		tag;
			Node	**child;
			int		nchild;
			Attr		**attr;
			int		nattr;
		};
		struct{				/* Ntext */
			Rune		*text;
		};
	};
	Node	*parent;
	Node	*next;
	Node	*prev;
};

enum
{
	Nelement,
	Ntext,
};

enum
{
	NTOK = 1024*8,
	NATTR = 512,
};

struct HtmlParser
{
	Rune		*name;
	int		lineno;
	int		tag;
	int		tagstart;
	int		pre;
	int		script;
	int		npeekc;
	Rune		tok[NTOK];
	Attr		attr[NATTR];
	int		nattr;

	Node 	*html;
	Node 	*head;
	Node 	*body;
	Node 	*current;

	int		eof;
	char		*in;
	int		nin;
	char		yysbuf[1024*8];
	char		*yysptr;
};

enum
{
	Ta,
	Tabbr,
	Tacronym,
	Taddress,
	Tapplet,
	Tarea,
	Tb,
	Tbase,
	Tbasefont,
	Tbdo,
	Tbig,
	Tblockquote,
	Tbody,
	Tbr,
	Tbutton,
	Tcanvas,
	Tcaption,
	Tcenter,
	Tcite,
	Tcode,
	Tcol,
	Tcolgroup,
	Tdd,
	Tdel,
	Tdfn,
	Tdir,
	Tdiv,
	Tdl,
	Tdt,
	Tem,
	Tembed,
	Tfieldset,
	Tfont,
	Tform,
	Tframe,
	Tframeset,
	Th1,
	Th2,
	Th3,
	Th4,
	Th5,
	Th6,
	Thead,
	Thr,
	Thtml,
	Ti,
	Tiframe,
	Timg,
	Tinput,
	Tins,
	Tisindex,
	Tkbd,
	Tkeygen,
	Tlabel,
	Tlayer,
	Tlegend,
	Tli,
	Tlink,
	Tmap,
	Tmarquee,
	Tmenu,
	Tmeta,
	Tnobr,
	Tnoembed,
	Tnoframes,
	Tnolayer,
	Tnoscript,
	Tobject,
	Tol,
	Toptgroup,
	Toption,
	Tp,
	Tparam,
	Tplaintext,
	Tpre,
	Tq,
	Ts,
	Tsamp,
	Tscript,
	Tselect,
	Tsmall,
	Tspacer,
	Tspan,
	Tstrike,
	Tstrong,
	Tstyle,
	Tsub,
	Tsup,
	Ttable,
	Ttbody,
	Ttd,
	Ttextarea,
	Ttfoot,
	Tth,
	Tthead,
	Ttitle,
	Ttr,
	Ttt,
	Tu,
	Tul,
	Tvar,
	Twbr,
	Txmp,
};

enum
{
	Aabbr,
	Aaccept,
	Aaccept_charset,
	Aaccesskey,
	Aaction,
	Aalign,
	Aalink,
	Aalt,
	Aarchive,
	Aautocomplete,
	Aautosave,
	Aaxis,
	Abackground,
	Abehavior,
	Abgcolor,
	Abgproperties,
	Aborder,
	Abordercolor,
	Acellborder,
	Acellpadding,
	Acellspacing,
	Achallenge,
	Achar,
	Acharoff,
	Acharset,
	Achecked,
	Acite,
	Aclass,
	Aclassid,
	Aclear,
	Acode,
	Acodebase,
	Acodetype,
	Acolor,
	Acols,
	Acolspan,
	Acompact,
	Acomposite,
	Acontent,
	Acontenteditable,
	Acoords,
	Adata,
	Adatetime,
	Adeclare,
	Adefer,
	Adir,
	Adirection,
	Adisabled,
	Aenctype,
	Aface,
	Afor,
	Aframe,
	Aframeborder,
	Aheaders,
	Aheight,
	Ahidden,
	Ahref,
	Ahreflang,
	Ahspace,
	Ahtml,
	Ahttp_equiv,
	Aid,
	Aincremental,
	Aismap,
	Akeytype,
	Alabel,
	Alang,
	Alanguage,
	Aleft,
	Aleftmargin,
	Alink,
	Alongdesc,
	Aloop,
	Amarginheight,
	Amarginwidth,
	Amax,
	Amaxlength,
	Amayscript,
	Amedia,
	Amethod,
	Amin,
	Amultiple,
	Aname,
	Anohref,
	Anoresize,
	Anosave,
	Anoshade,
	Anowrap,
	Aobject,
	Aonabort,
	Aonbeforecopy,
	Aonbeforecut,
	Aonbeforepaste,
	Aonblur,
	Aonchange,
	Aonclick,
	Aoncontextmenu,
	Aoncopy,
	Aoncut,
	Aondblclick,
	Aondrag,
	Aondragend,
	Aondragenter,
	Aondragleave,
	Aondragover,
	Aondragstart,
	Aondrop,
	Aonerror,
	Aonfocus,
	Aoninput,
	Aonkeydown,
	Aonkeypress,
	Aonkeyup,
	Aonload,
	Aonmousedown,
	Aonmousemove,
	Aonmouseout,
	Aonmouseover,
	Aonmouseup,
	Aonpaste,
	Aonreset,
	Aonresize,
	Aonscroll,
	Aonsearch,
	Aonselect,
	Aonselectstart,
	Aonsubmit,
	Aonunload,
	Aoversrc,
	Apagex,
	Apagey,
	Aplaceholder,
	Aplain,
	Apluginpage,
	Apluginspage,
	Apluginurl,
	Aprecision,
	Aprofile,
	Aprompt,
	Areadonly,
	Arel,
	Aresults,
	Arev,
	Arows,
	Arowspan,
	Arules,
	Ascheme,
	Ascope,
	Ascrollamount,
	Ascrolldelay,
	Ascrolling,
	Aselected,
	Ashape,
	Asize,
	Aspan,
	Asrc,
	Astandby,
	Astart,
	Astyle,
	Asummary,
	Atabindex,
	Atableborder,
	Atarget,
	Atext,
	Atitle,
	Atop,
	Atopmargin,
	Atruespeed,
	Atype,
	Aunknown,
	Ausemap,
	Avalign,
	Avalue,
	Avaluetype,
	Aversion,
	Avisibility,
	Avlink,
	Avspace,
	Awidth,
	Awrap,
};

extern	void	htmlinit(HtmlParser *, Rune *);
extern	void	htmlparse(HtmlParser *, char *, int);
extern	void	htmlprtree(Node *);
extern	void	htmlprtext(Node *);

extern	Node*	allocnode(int, int);
extern	void	freenode(Node *, int);
extern	void	addnode(Node *, Node *);
extern	void	rmnode(Node *);
extern	void mvnode(Node *, Node *);

extern	Rune*	getattr(Node *, Rune *);

extern	int	htmldebug[];

#pragma	varargck	type	"N"	Node*

extern	void 	*emalloc(ulong);
extern	void	*erealloc(void *, ulong);

