#include <u.h>
#include <libc.h>
#include "html.h"

Node *
allocnode(int type, int tag)
{
	Node *n;

	n = emalloc(sizeof(Node));
	memset(n, 0, sizeof(*n));
	n->type = type;
	n->tag = tag;
	n->id = -1;
	n->nchild = 0;
	return n;
}

void
freenode(Node *n, int recursive)
{
	int i;

	if(n == nil)
		return;

	switch(n->type){

	case Ntext:
		free(n->text);
		break;

	case Nelement:
		for(i=0; i<n->nattr; i++){
			free(n->attr[i]->name);
			free(n->attr[i]->value);
			free(n->attr[i]);
		}
		free(n->attr);
		if(recursive)
			for(i=0; i<n->nchild; i++)
				freenode(n->child[i], 1);
		free(n->child);
		break;
	}
	free(n);
}

void
addnode(Node *p, Node *n)
{
	n->parent = p;
	n->id = p->nchild;
	p->child = erealloc(p->child, ++p->nchild*sizeof(Node));
	p->child[n->id] = n;
	if(n->id){
		n->prev = p->child[n->id-1];
		n->prev->next = n;
	}
}

void
rmnode(Node *n)
{
	int i;

	assert(n != nil);
	assert(n->parent != nil);
	assert(n->parent->child != nil);
	for(i=n->id; i<n->parent->nchild-1; i++)
		n->parent->child[i] = n->parent->child[i+1];
	if(n->prev)
		n->prev->next = n->next;
	if(n->next)
		n->next->prev = n->prev;
	n->parent->nchild--;
}

void
mvnode(Node *n1, Node *n2)
{
	assert(n1 != nil);
	assert(n2 != nil);
	assert(n1->parent != nil);
	assert(n1->parent->child != nil);
	n2->id = n1->id;
	n2->parent = n1->parent;
	n2->parent->child[n1->id] = n2;
	if(n1->prev){
		n1->prev->next = n2;
		n2->prev = n1->prev;
	}
	if(n1->next){
		n1->next->prev = n2;
		n2->next = n1->next;
	}
}

Rune *
getattr(Node *n, Rune *name)
{
	int i;

	for(i=0; i<n->nattr; i++)
		if(runestrcmp(n->attr[i]->name, name) == 0)
			return n->attr[i]->value;

	return nil;
}
